"""Top-level package."""

__author__ = (
    """Yann Golhen, Yann Lagré, Sebastien Bidault, Maxime Gendre, Thomas Bouche, Johann Martin, Guillaume Vignal"""
)
__email__ = "yann.golhen@maif.fr, yann.lagre@maif.fr, sebabstien.bidault.marketing@maif.fr, thomas.bouche@maif.fr, guillaume.vignal@maif.fr"

from shapash.explainer.smart_explainer import SmartExplainer

from .__version__ import __version__
